#include "volume.h"
#include "vvdebugmsg.h"

#define DFT_BRICK_SIZE 128
#define FILENAME_SIZE 512

Volume::Volume() {
	vd = NULL; 

	setBrickSize(DFT_BRICK_SIZE);
	setVolOrganization(VOL_ORG_MIX);

	fpVol = NULL; 
	whereData = 0;
}

Volume::~Volume() {
	if(vd != NULL) delete vd;
	if(fpVol != NULL) fclose(fpVol);
}

void Volume::setBrickSize(int bS) {
	brickSize = bS;

	if(vd != NULL) {
		for(int i=0; i<3; i++) {
			nBricks[i] = vd->vox[i]/brickSize;
			remaingPixels[i] = vd->vox[i]%brickSize;
			if(remaingPixels[i] != 0) nBricks[i]++;
		}
		sliceBytes = vd->getSliceBytes();
		BPV = vd->getBPV();
	} else {
		nBricks[0] =  nBricks[1] =  nBricks[2] = 0;
		remaingPixels[0] = remaingPixels[1] = remaingPixels[2] = 0;
		sliceBytes = 0;
		BPV = 0;
	}
}

int Volume::getBrickSize() {
	return brickSize;
}

void Volume::setVolOrganization(unsigned int volOrg) {
	volOrganization = volOrg;
}

// get single channel data from a file
bool Volume::getOneChannelData(int x, int y, int z, int c, unsigned char *raw) {
	switch(volOrganization) {
		case VOL_ORG_SEP:
			return getOneChannelData_sep(x, y, z, c, raw);
			break;
		case VOL_ORG_MIX:
			return getOneChannelData_mix(x, y, z, c, raw);
			break;
	}
	return false;
}

// get multiple channel data from a file
bool Volume::getMultiChannelData(int x, int y, int z, unsigned char *raw) {
	switch(volOrganization) {
		case VOL_ORG_SEP:
			return getMultiChannelData_sep(x, y, z, raw);
			break;
		case VOL_ORG_MIX:
			return getMultiChannelData_mix(x, y, z, raw);
			break;
	}
	return false;
}

// get one channel data from a file with RGB saved separately (sep)
bool Volume::getOneChannelData_sep(int x, int y, int z, int c, uchar *raw) {
	vvDebugMsg::msg( 2, "==getOneChannelData_sep()==");

	if( fpVol == NULL) {
		vvDebugMsg::msg( 1, "Error: file not opened.\n");
		return -1;
	}

	if( c>vd->chan ) {
		vvDebugMsg::msg( 1, "Error: this channel does not exist. # channels: ", c, vd->chan);
		return -1;
	}

	if(x >= nBricks[0] || y >= nBricks[1] || z >= nBricks[2] ||
	    x < 0 || y < 0 || z < 0) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - out of range.");
		return -1;
	}

	int rawSize[3] = { (x == nBricks[0]-1 && remaingPixels[0] != 0)? remaingPixels[0]: brickSize,
			   (y == nBricks[1]-1 && remaingPixels[1] != 0)? remaingPixels[1]: brickSize,
			   (z == nBricks[2]-1 && remaingPixels[2] != 0)? remaingPixels[2]: brickSize };

	int startPtr = whereData + BPV*brickSize*(x*vd->vox[1]*vd->vox[2] + rawSize[0]*y*vd->vox[2] + rawSize[0]*rawSize[1]*z);
	int readBytes = rawSize[0]*rawSize[1]*rawSize[2]*vd->bpc;

	fseek(fpVol, startPtr+c*readBytes, SEEK_SET);
	if ((int)fread(raw, 1, readBytes, fpVol) != readBytes) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - read brick data.");
		vvDebugMsg::msg(2, "                                     - readBytes: ", readBytes);
		vvDebugMsg::msg(2, "                                     - startPtr: ", startPtr);
		return false;
	}
	return true;

}

// not checked yet
// get one channel data from a file with RGB saved separately (sep)
bool Volume::getOneChannelData_mix(int x, int y, int z, int c, uchar *raw) {
	vvDebugMsg::msg( 2, "==getOneChannelData_mix()==");

	if( fpVol == NULL) {
		vvDebugMsg::msg( 1, "Error: file not opened.\n");
		return -1;
	}

	if( c>vd->chan ) {
		vvDebugMsg::msg( 1, "Error: this channel does not exist. # channels: ", c, vd->chan);
		return -1;
	}

	if(x >= nBricks[0] || y >= nBricks[1] || z >= nBricks[2] ||
	    x < 0 || y < 0 || z < 0) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - out of range.");
		return -1;
	}

	int rawSize[3] = { (x == nBricks[0]-1 && remaingPixels[0] != 0)? remaingPixels[0]: brickSize,
			   (y == nBricks[1]-1 && remaingPixels[1] != 0)? remaingPixels[1]: brickSize,
			   (z == nBricks[2]-1 && remaingPixels[2] != 0)? remaingPixels[2]: brickSize };

	int startPtr = whereData + BPV*brickSize*(x*vd->vox[1]*vd->vox[2] + rawSize[0]*y*vd->vox[2] + rawSize[0]*rawSize[1]*z);
	int readBytes = rawSize[0]*rawSize[1]*rawSize[2]*vd->bpc;
        int readBytes_nC = readBytes*vd->chan;

	uchar *bufAllChannel = new uchar [readBytes_nC];
	memset( raw, 0, readBytes );
	memset( bufAllChannel, 0, readBytes_nC);

	fseek(fpVol, startPtr, SEEK_SET);
	if ((int)fread(bufAllChannel, 1, readBytes_nC, fpVol) != readBytes_nC) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - read brick data.");
		vvDebugMsg::msg(2, "                                     - readBytes: ", readBytes);
		vvDebugMsg::msg(2, "                                     - startPtr: ", startPtr);
		return false;
	}

	// reorganizae the data
	for(int i=c*vd->bpc, j=0; i<readBytes_nC; i+=BPV, j+=vd->bpc)
		for(int p=0; p<vd->bpc; p++)
			raw[j+p] = bufAllChannel[i+p];

	delete [] bufAllChannel;
	return true;

}

// get a brick in our file format
bool Volume::getMultiChannelData_mix(int x, int y, int z, uchar *raw) {
	if(x >= nBricks[0] || y >= nBricks[1] || z >= nBricks[2] ||
	   x < 0 || y < 0 || z < 0) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - out of range.");
		return false;
	}
	if(raw == NULL) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - raw = NULL.");
		return false;
	}

	int rawSize[3] = { (x == nBricks[0]-1 && remaingPixels[0] != 0)? remaingPixels[0]: brickSize,
			   (y == nBricks[1]-1 && remaingPixels[1] != 0)? remaingPixels[1]: brickSize,
			   (z == nBricks[2]-1 && remaingPixels[2] != 0)? remaingPixels[2]: brickSize };

	int startPtr = whereData + BPV*brickSize*( x*vd->vox[1]*vd->vox[2] + rawSize[0]*y*vd->vox[2] + rawSize[0]*rawSize[1]*z);
	int readBytes = rawSize[0]*rawSize[1]*rawSize[2]*BPV;

	fseek(fpVol, startPtr, SEEK_SET);
	if ((int)fread(raw, 1, readBytes, fpVol) != readBytes) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - read brick data.");
		vvDebugMsg::msg(2, "                                     - readBytes: ", readBytes);
		vvDebugMsg::msg(2, "                                     - startPtr: ", startPtr);
		return false;
	}
	return true;
}
		
// not checked yet
// get a brick in our file format
bool Volume::getMultiChannelData_sep(int x, int y, int z, uchar *raw) {
	if(x >= nBricks[0] || y >= nBricks[1] || z >= nBricks[2] ||
	   x < 0 || y < 0 || z < 0) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - out of range.");
		return false;
	}
	if(raw == NULL) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - raw = NULL.");
		return false;
	}

	int rawSize[3] = { (x == nBricks[0]-1 && remaingPixels[0] != 0)? remaingPixels[0]: brickSize,
			   (y == nBricks[1]-1 && remaingPixels[1] != 0)? remaingPixels[1]: brickSize,
			   (z == nBricks[2]-1 && remaingPixels[2] != 0)? remaingPixels[2]: brickSize };

	int startPtr = whereData + BPV*brickSize*( x*vd->vox[1]*vd->vox[2] + rawSize[0]*y*vd->vox[2] + rawSize[0]*rawSize[1]*z);
	int readBytes = rawSize[0]*rawSize[1]*rawSize[2]*vd->bpc;
	int readBytes_nC = readBytes * vd->chan;
        uchar *bufAllChannel = new uchar [readBytes_nC];
        memset( raw, 0, readBytes_nC );
        memset( bufAllChannel, 0, readBytes_nC);

	fseek(fpVol, startPtr, SEEK_SET);
	if ((int)fread(bufAllChannel, 1, readBytes_nC, fpVol) != readBytes_nC) {
		vvDebugMsg::msg(1, "Error: Volume::getMultiChannelData() - read brick data.");
		vvDebugMsg::msg(2, "                                     - readBytes: ", readBytes);
		vvDebugMsg::msg(2, "                                     - startPtr: ", startPtr);
		return false;
	}

	// reorganize the data
	for(int c=0; c<vd->chan; c++)
		for(int j=c*vd->bpc, i=c*readBytes; j<readBytes_nC; j+=BPV, i+=vd->bpc)
			for(int p=0; p<vd->bpc; p++)
				raw[j+p] = bufAllChannel[i+p];

	delete [] bufAllChannel;

	return true;
}
